/* 
 * glDino - 3D motor to render three-dimentional objects
 *
 * Mesa 3D graphics library 2.6 version
 * Linux 2.0.x or higher
 *
 * Author: Navarro Mikael (klnavarro@multimania.com)
 *
 */

/*
 * ALL RIGHTS RESERVED 
 * Permission to use, copy, modify, and distribute this software for 
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice appear in all copies and that both the copyright notice
 * and this permission notice appear in supporting documentation, and that 
 * the name of OpenGL Inc. not be used in advertising or publicity pertaining
 * to distribution of the software without specific, written prior permission.
 *
 * THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU "AS-IS"
 * AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR OTHERWISE,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
 * FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL SILICON
 * GRAPHICS, INC.  BE LIABLE TO YOU OR ANYONE ELSE FOR ANY DIRECT,
 * SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY
 * KIND, OR ANY DAMAGES WHATSOEVER, INCLUDING WITHOUT LIMITATION,
 * LOSS OF PROFIT, LOSS OF USE, SAVINGS OR REVENUE, OR THE CLAIMS OF
 * THIRD PARTIES, WHETHER OR NOT SILICON GRAPHICS, INC.  HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH LOSS, HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * POSSESSION, USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/* Compile: gcc -ansi -O2 -o glDino glDino.c dino.c trackball.c 
   -I/usr/include/gtk-1.2 -I/usr/include/glib-1.2 -I/usr/lib/glib/include 
   -L/usr/X11R6/lib -lgtkgl -lGLU -lGL -lm */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> /* for cos(), sin(), and sqrt() */
#include <GL/gl.h>
#include <GL/glu.h> /* OpenGL Utility Library */
#include <gtk/gtk.h>
#include <gtkgl/gtkglarea.h>

#include "dino.h" /* draw Dinosaur */
#include "trackball.h" /* virtual trackball */

/* Some <math.h> files do not define M_PI... */
#ifndef M_PI
#define M_PI 3.14159265359
#endif

/* Some usefull macros */
#define MIN_VALUE(a,b) (((a) < (b)) ? a : b)
#define MAX_VALUE(a,b) (((a) > (b)) ? a : b)
#define RAD2DEG(x) (x * 180 / M_PI)
#define DEG2RAD(x) (x * M_PI / 180)


#define VIEW_ASPECT 1.3

char help_text[] = "glDino  :-)\n"
                   "Usage: gltrack [OPTION]... [FILE]...\n"
                   "Options:\n"
                   "  --help   display this message\n"
                   "Mouse:\n"
                   "  Mouse button 1 + drag   spin (virtual trackball)\n"
                   "  Mouse button 2 + drag   zoom\n"
                   "  Mouse button 3          popup menu\n"
                   "Keys:\n"
                   "  <ESC> quit\n"
                   "\n";

/* Information needed to display lightwave mesh */
typedef struct {
  float startx, starty; /* position of mouse */
  float zoom;           /* scale factor of object */
  float quat[4];        /* orientation of object */
} mesh_info;

/* Forward declaration */
static void glarea_initgl(GtkWidget *widget);

static gint glarea_display(GtkWidget *widget, GdkEventExpose *event);
static gint glarea_reshape(GtkWidget *widget, GdkEventConfigure *event);
static gint glarea_destroy(GtkWidget *widget);
static gint glarea_mouse(GtkWidget *widget, GdkEventButton *event);
static gint glarea_motion(GtkWidget *widget, GdkEventMotion *event);

static gint popup_menu_handler(GtkWidget *widget, GdkEventButton *event);
static void popup_menu_detacher(GtkWidget *attach_widget, GtkMenu *menu);
static void create_popup_menu(GtkWidget *widget);

static gint glarea_gouraud_shading(GtkWidget *widget, GtkObject *object);
static gint glarea_flat_shading(GtkWidget *widget, GtkObject *object);
static gint glarea_wireframe(GtkWidget *widget, GtkObject *object);
static gint glarea_hidden_lines(GtkWidget *widget, GtkObject *object);
static gint glarea_only_points(GtkWidget *widget, GtkObject *object);

static gint window_destroy(GtkWidget *widget);
static gint glarea_draw(void);


static void glarea_initgl(GtkWidget *widget) {
  GLfloat light0_position[] = {-10.0, 10.0, -10.0, 0.0}; /* sun */
  GLfloat light0_ambiant[] = {0.6, 0.6, 0.6, 1.0};
  GLfloat light0_diffuse[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat light0_specular[] = {1.0, 1.0, 1.0, 2.0};
  GLfloat light1_position[] = {-10.0, 10.0, 10.0, 1.0}; 
  GLfloat light2_position[] = {10.0, 10.0, 10.0, 1.0}; 
  
  if(gtk_gl_area_make_current(GTK_GL_AREA(widget))) {
    /* Remove back faces */
    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
  
    /* Speedups */
    glEnable(GL_DITHER);
    glShadeModel(GL_SMOOTH);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

    /* View */
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glMatrixMode(GL_PROJECTION);
    gluPerspective(40.0f, VIEW_ASPECT, 1.0f, 100.0f);
    glMatrixMode(GL_MODELVIEW);
    gluLookAt(0.0f, 10.0f, 50.0f,
	      0.0f, 0.0f, 0.0f,
	      0.0f, 1.0f, 0.0f);
    glEnable(GL_NORMALIZE); /* In case we're scaling */

    glPushMatrix(); /* dummy push so we can pop on model recalc */

    /* Light */
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_diffuse);  
    glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.1f);
    glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.05f);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
    glLightfv(GL_LIGHT1, GL_DIFFUSE,  light0_diffuse);  
    glLightfv(GL_LIGHT1, GL_SPECULAR, light0_specular);
    glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0.1f);
    glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.05f);
    glEnable(GL_LIGHT1);
    glLightfv(GL_LIGHT2, GL_POSITION, light2_position);
    glLightfv(GL_LIGHT2, GL_DIFFUSE,  light0_diffuse);  
    glLightfv(GL_LIGHT2, GL_SPECULAR, light0_specular);
    glLightf(GL_LIGHT2, GL_CONSTANT_ATTENUATION, 0.1f);
    glLightf(GL_LIGHT2, GL_LINEAR_ATTENUATION, 0.05f);
    glEnable(GL_LIGHT2);
    glEnable(GL_LIGHTING);
  }
}

static gint glarea_display(GtkWidget *widget, GdkEventExpose *event) {
  GLfloat m[4][4];
  GtkGLArea *glarea = GTK_GL_AREA(widget);
  mesh_info *info = (mesh_info*)gtk_object_get_data(GTK_OBJECT(widget), 
						    "mesh_info");

  /* Draw only last expose */
  if(event->count > 0) {
    return TRUE;
  }

  /* OpenGL calls can be done only if make_current returns true */
  if(gtk_gl_area_make_current(glarea)) {
    /* recalcModelView */
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glPushMatrix();
    build_rotmatrix(m, info->quat);
    glMultMatrixf(&m[0][0]);
    glScalef(info->zoom, info->zoom, info->zoom);
     
    /* Draw object */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glTranslatef(0, -8, 0); /* translate the dinosaur to be at (0,8,0) */
    drawDinosaur();

    /* Swap backbuffer to front */
    gtk_gl_area_swapbuffers(glarea);
  }

  return TRUE;
}

static gint glarea_reshape(GtkWidget *widget, GdkEventConfigure *event) {
  /* OpenGL calls can be done only if make_current returns true */
  if(gtk_gl_area_make_current(GTK_GL_AREA(widget))) {
    glViewport(0, 0, widget->allocation.width, widget->allocation.height);
  }
  return TRUE;
}

static gint glarea_destroy(GtkWidget *widget) {
  /* Delete mesh info */
  mesh_info *info = (mesh_info*)gtk_object_get_data(GTK_OBJECT(widget),
						    "mesh_info");

  if(info) {
    g_free(info);
  }
  return TRUE;
}

static gint glarea_mouse(GtkWidget *widget, GdkEventButton *event) {
  mesh_info *info = (mesh_info*)gtk_object_get_data(GTK_OBJECT(widget), 
						    "mesh_info");

  if(event->button == 1) {
    info->startx = event->x;
    info->starty = event->y;
    return TRUE;
  }
  return FALSE;
}

static gint glarea_motion(GtkWidget *widget, GdkEventMotion *event) {
  int x, y;
  GdkRectangle area;
  GdkModifierType state;
  mesh_info *info = (mesh_info*)gtk_object_get_data(GTK_OBJECT(widget), 
						    "mesh_info");

  if (event->is_hint) {
#if !defined(WIN32) /* fix this! */
    gdk_window_get_pointer(event->window, &x, &y, &state);
#endif
  } else {
    x = event->x;
    y = event->y;
    state = event->state;
  }
  
  area.x = 0;
  area.y = 0;
  area.width  = widget->allocation.width;
  area.height = widget->allocation.height;

  if (state & GDK_BUTTON1_MASK) {
    /* Drag in progress, simulate trackball */
    float spin_quat[4];
    trackball(spin_quat,
	      (2.0*info->startx - area.width) / area.width,
	      (area.height - 2.0*info->starty) / area.height,
	      (2.0*x - area.width) / area.width,
	      (area.height - 2.0*y) / area.height);
    add_quats(spin_quat, info->quat, info->quat);

    /* Orientation has changed, redraw mesh */
    gtk_widget_draw(widget, &area);
  }

  if (state & GDK_BUTTON2_MASK) {
    /* Zooming drag */
    info->zoom = info->zoom * (1.0f + (info->starty - y) / area.height);

    /* Zoom has changed, redraw mesh */
    gtk_widget_draw(widget, &area);
  }
  info->startx = x;
  info->starty = y;

  return TRUE;
}

static gint popup_menu_handler(GtkWidget *widget, GdkEventButton *event) {
  if (event->button == 3) {
    gtk_menu_popup(GTK_MENU(widget), NULL, NULL, NULL, NULL,
		   event->button, event->time);
    return TRUE;
  }
  return FALSE;
}
static void popup_menu_detacher(GtkWidget *attach_widget, GtkMenu *menu) {
}
static void create_popup_menu(GtkWidget *widget) {
  GtkWidget *menu          = gtk_menu_new();
  GtkWidget *open_item     = gtk_menu_item_new_with_label("Open");
  GtkWidget *quit_item     = gtk_menu_item_new_with_label("Quit");
  GtkWidget *quit_all_item = gtk_menu_item_new_with_label("Quit All");

  gtk_menu_append(GTK_MENU(menu), open_item);
  gtk_menu_append(GTK_MENU(menu), quit_item);
  gtk_menu_append(GTK_MENU(menu), quit_all_item);

  gtk_widget_show(open_item);
  gtk_widget_show(quit_item);
  gtk_widget_show(quit_all_item);

  gtk_menu_attach_to_widget(GTK_MENU(menu), GTK_WIDGET(widget),
			    popup_menu_detacher);
  gtk_signal_connect_object(GTK_OBJECT(widget), "destroy",
			    GTK_SIGNAL_FUNC(gtk_menu_detach), 
			    GTK_OBJECT(menu));
  gtk_signal_connect_object(GTK_OBJECT(widget), "button_press_event",
			    GTK_SIGNAL_FUNC(popup_menu_handler), 
			    GTK_OBJECT(menu));

  gtk_signal_connect_object(GTK_OBJECT(open_item), "activate",
			    GTK_SIGNAL_FUNC(glarea_draw), 
			    GTK_OBJECT(widget));
  gtk_signal_connect_object(GTK_OBJECT(quit_item), "activate",
			    GTK_SIGNAL_FUNC(gtk_widget_destroy), 
			    GTK_OBJECT(widget));
  gtk_signal_connect(GTK_OBJECT(quit_all_item), "activate",
		     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
}

static gint glarea_gouraud_shading(GtkWidget *widget, GtkObject *object) {
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
    if(gtk_gl_area_make_current(GTK_GL_AREA(object))) {
      glEnable(GL_CULL_FACE);
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      glShadeModel(GL_SMOOTH);
    }
    gtk_widget_queue_draw(GTK_WIDGET(object));
  }
}
static gint glarea_flat_shading(GtkWidget *widget, GtkObject *object) {
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
    if(gtk_gl_area_make_current(GTK_GL_AREA(object))) {
      glEnable(GL_CULL_FACE);
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      glShadeModel(GL_FLAT);
    }
    gtk_widget_queue_draw(GTK_WIDGET(object));
  }
}
static gint glarea_wireframe(GtkWidget *widget, GtkObject *object) {
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
    if(gtk_gl_area_make_current(GTK_GL_AREA(object))) {
      glDisable(GL_CULL_FACE);
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      glShadeModel(GL_SMOOTH);
    }
    gtk_widget_queue_draw(GTK_WIDGET(object));
  }
}
static gint glarea_hidden_lines(GtkWidget *widget, GtkObject *object) {
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
    if(gtk_gl_area_make_current(GTK_GL_AREA(object))) {
      glEnable(GL_CULL_FACE);
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      glShadeModel(GL_SMOOTH);
    }
    gtk_widget_queue_draw(GTK_WIDGET(object));
  }
}
static gint glarea_only_points(GtkWidget *widget, GtkObject *object) {
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
    if(gtk_gl_area_make_current(GTK_GL_AREA(object))) {
      glDisable(GL_CULL_FACE);
      glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
      glShadeModel(GL_SMOOTH);
    }
    gtk_widget_queue_draw(GTK_WIDGET(object));
  }
}

static gint window_count = 0; /* number of windows on screen */

static gint window_destroy(GtkWidget *widget) {
  /* If this was last window quit */
  if (--window_count == 0)
    gtk_main_quit();
  return TRUE;
}

static gint glarea_draw(void) {
  GtkWidget *window, *glarea;
  GtkWidget *hbox, *vbox, *frame, *vbox2, *button;
  GSList *group;
  mesh_info *info;

  /* Read object */

  /* Create new top level window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "glTrack");
  gtk_container_set_border_width(GTK_CONTAINER(window), 10);
  create_popup_menu(window); /* add popup menu to window */
  gtk_signal_connect (GTK_OBJECT(window), "destroy",
		      GTK_SIGNAL_FUNC(window_destroy), NULL);
  window_count++;

  /* Destroy this window when exiting from gtk_main() */
  gtk_quit_add_destroy(1, GTK_OBJECT(window));

  /* Horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), hbox);

  /* Create new OpenGL widget */
  glarea = gtk_gl_area_new_vargs(NULL, /* no sharing */
				 GDK_GL_RGBA,
				 GDK_GL_RED_SIZE,1,
				 GDK_GL_GREEN_SIZE,1,
				 GDK_GL_BLUE_SIZE,1,
				 GDK_GL_DEPTH_SIZE,1,
				 GDK_GL_DOUBLEBUFFER,
				 GDK_GL_NONE); /* last arg GDK_GL_NONE */
  if (glarea == NULL) {
    g_print("Can't create GtkGLArea widget\n");
    return FALSE;
  }

  /* Set up events and signals for OpenGL widget */
  gtk_widget_set_events(glarea,
			GDK_EXPOSURE_MASK|
			GDK_BUTTON_PRESS_MASK|
			GDK_BUTTON_RELEASE_MASK|
			GDK_POINTER_MOTION_MASK);
  gtk_signal_connect(GTK_OBJECT(glarea), "realize",
		     GTK_SIGNAL_FUNC(glarea_initgl), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "expose_event",
		     GTK_SIGNAL_FUNC(glarea_display), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "configure_event",
		     GTK_SIGNAL_FUNC(glarea_reshape), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "motion_notify_event",
		     GTK_SIGNAL_FUNC(glarea_motion), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "button_press_event",
		     GTK_SIGNAL_FUNC(glarea_mouse), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "destroy",
		     GTK_SIGNAL_FUNC (glarea_destroy), NULL);

  gtk_widget_set_usize(glarea, 512, 512/VIEW_ASPECT); /* minimum size */

  /* Set up mesh info */
  info = (mesh_info*)g_malloc(sizeof(mesh_info));
  info->startx = 0;
  info->starty = 0;
  info->zoom = 1;
  trackball(info->quat , 0.0, 0.0, 0.0, 0.0);
  gtk_object_set_data(GTK_OBJECT(glarea), "mesh_info", info);

  /* Put glarea into window */ 
  gtk_box_pack_start(GTK_BOX(hbox), glarea, TRUE, TRUE, 0);

  /* Vertical box : Controls */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 10);

  /* Frame : Render Mode */
  frame = gtk_frame_new("Render Mode");
  gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);

  /* Vertical box : radio buttons */
  vbox2 = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(frame), vbox2);

  /* Grp radio */
  button = gtk_radio_button_new_with_label(NULL, "Gouraud Shading");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
  gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
  gtk_signal_connect(GTK_OBJECT(button), "toggled",
		     GTK_SIGNAL_FUNC(glarea_gouraud_shading), 
		     GTK_OBJECT(glarea));
  group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));

  button = gtk_radio_button_new_with_label(group, "Flat Shading");
  gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
  gtk_signal_connect(GTK_OBJECT(button), "toggled",
		     GTK_SIGNAL_FUNC(glarea_flat_shading), 
		     GTK_OBJECT(glarea));
  group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));

  button = gtk_radio_button_new_with_label(group, "Wireframe");
  gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
  gtk_signal_connect(GTK_OBJECT(button), "toggled",
		     GTK_SIGNAL_FUNC(glarea_wireframe), 
		     GTK_OBJECT(glarea));
  group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));

  button = gtk_radio_button_new_with_label(group, "Hidden Lines");
  gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
  gtk_signal_connect(GTK_OBJECT(button), "toggled",
		     GTK_SIGNAL_FUNC(glarea_hidden_lines), 
		     GTK_OBJECT(glarea));
  group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));

  button = gtk_radio_button_new_with_label(group, "Only Points");
  gtk_signal_connect(GTK_OBJECT(button), "toggled",
		     GTK_SIGNAL_FUNC(glarea_only_points), 
		     GTK_OBJECT(glarea));
  gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);

  /* Button : draw Dinosaur */
  button = gtk_check_button_new_with_label("Toggle Dinosaur");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

  /* Button : Lights */
  button = gtk_check_button_new_with_label("Toggle Lights");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

  /* Button : Quit */
  button = gtk_button_new_with_label("Quit");
  gtk_box_pack_end(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			    GTK_SIGNAL_FUNC(gtk_widget_destroy), 
			    GTK_OBJECT(window));

  /* And show all */
  /*gtk_widget_show(glarea);*/
  gtk_widget_show_all(window);

  return TRUE;
}

int main(int argc, char **argv) {
  int i;

  /* Initialize gtk */
  gtk_init(&argc, &argv);

   /* Check if OpenGL (GLX extension) is supported */
  if (gdk_gl_query() == FALSE) {
    g_print("OpenGL not supported\n");
    return 0;
  }

  /* Help ? */
  for(i=1; i<argc; i++) {
    if(!strcmp(argv[i],"--help") || !strcmp(argv[i],"-h")) {
      g_print("%s", help_text);
      exit(0);
    }
  }

  if(argc == 1) {
    /* no filenames, show filerequester */
  } else {
    /* show requested files */
  }

  glarea_draw(); /* draw interface */

  if(window_count > 0)
    gtk_main();

  return 0;
}
